#include "engine.h"
#include "texturesdl.h"
#include "sdl_opengl.h"

using std::string;
using namespace EPOC;

/** Default constructor. **/
Engine::Engine()
{
	SetClassID(kCLASSID_ENGINE);

	m_pScreen = NULL;
	m_sTitle = "";
 
	m_lLastTick = 0;
	m_nFPSTickCounter = 0;
	m_nFPSCounter = 0;
	m_nCurrentFPS = 50;
	m_nBpp = 0;

	m_fTimeDelta = 0.0f;
 
	m_bMinimized = false;
	m_bFullScreen = false;
	m_bAudioSupport = true;

	m_nWidth = 0;
	m_nHeight = 0;
	m_nFlags = 0;
	m_nRenderMode = kRENDER_MODE_UNKNOWN;
	m_pVideoInfo = NULL;

	SetFPSCapEnabled(false);
	SetFPSCap(60);

	m_nQuitKey = SDLK_ESCAPE;
}
 
/** Destructor. **/
Engine::~Engine()
{
	Engine::Cleanup();
}

void Engine::Cleanup()
{
	if(m_pScreen)
	{
		m_pScreen->Release();
		m_pScreen = NULL;
	}

	TTF_Quit();
	SDL_Quit();
}


//Initialize SDL, the window and the additional data
void Engine::Start(string sTitle, int nWidth, int nHeight, bool bFullScreen, RENDER_MODE nRenderMode, int nCusomFlags)
{

	printf( "Engine: Starting up the engine!\n");

	// register SDL_Quit to be called at exit; makes sure things are cleaned up when we quit.
	atexit(SDL_Quit );

	// initialize SDL's subsystems
	if(SDL_Init(SDL_INIT_VIDEO | SDL_INIT_TIMER) < 0) 
	{
		fprintf( stderr, "Engine: Error, unable to initialize SDL: %s\n", SDL_GetError() );
		exit( 1 );
	}

	if(SDL_InitSubSystem(SDL_INIT_AUDIO) < 0)
	{
		fprintf( stderr, "Engine: Error, unable to initialize SDL subsystem AUDIO: %s\n", SDL_GetError() );
		fprintf( stderr, "Engine: Error, audio disabled\n");
		//exit( 1 );
		m_bAudioSupport = false;
	}

	printf( "Engine: SDL was successfully initialized\n");

	m_nRenderMode = nRenderMode;
	if(m_nRenderMode != kRENDER_MODE_SDL && m_nRenderMode != kRENDER_MODE_OPENGL)
	{
		fprintf( stderr, "Engine: Error, render mode undefined\n");
		exit(1);
	}
	
	m_pVideoInfo = SDL_GetVideoInfo( );
	
	switch(m_nRenderMode)
	{
		case kRENDER_MODE_SDL:
		{
			if(nCusomFlags > -1 )
				m_nFlags = nCusomFlags;
			else
				m_nFlags = SDL_HWSURFACE | SDL_DOUBLEBUF | SDL_HWPALETTE;
		}
		break;

		case kRENDER_MODE_OPENGL:
		{
			m_nFlags = SDL_OPENGL;
			/*	SDL_GL_SetAttribute( SDL_GL_RED_SIZE, 5 );		
			SDL_GL_SetAttribute( SDL_GL_GREEN_SIZE, 5 );	
			SDL_GL_SetAttribute( SDL_GL_BLUE_SIZE, 5 );
			SDL_GL_SetAttribute( SDL_GL_DEPTH_SIZE, 16 );*/

			SDL_GL_SetAttribute( SDL_GL_DOUBLEBUFFER, 1 );
		}
		break;
	}
		


	// Attempt to create a window with the specified height and width.
	if(SetVideoMode(nWidth, nHeight, bFullScreen, 32, m_nFlags) == false)
	{
		fprintf( stderr, "Engine: Error, unable to setup video mode (%s)\n", SDL_GetError() );
		exit( 1 );
	}
	
	if(TTF_Init()==-1)
	{
		printf("Engine: Error, unable to initialize TTF (%s)\n", TTF_GetError());
		exit(1);
	}

#ifdef _DEBUG
	SetTitle(DEFAULT_TITLE + sTitle);
#else
	SetTitle(sTitle);
#endif

	if(Initialize() == false)
	{
		fprintf( stderr, "Engine: Error, failed to initialize engine\n");
		exit(1);
	}

	m_lLastTick = SDL_GetTicks();
	m_bQuit = false;
 
	// Main loop: loop forever.
	while(!m_bQuit)
	{
		// Handle events
		DoEvents();

		if(DoUpdate() == false)
			m_bQuit = true;
 
		// render stuff
		DoRender();
	}
}
 
/** Handles all controller inputs.
	@remark This function is called once per frame.
**/
void Engine::DoEvents()
{
	// Poll for events, and handle the ones we care about.
	SDL_Event event;
	while(SDL_PollEvent(&event)) 
	{
		// Handle events in children
		HandleEvent(&event);

		switch(event.type ) 
		{
			case SDL_KEYDOWN:
			{
				// If escape is pressed set the Quit-flag
				if (event.key.keysym.sym == m_nQuitKey)
				{
					m_bQuit = true;
					break;
				}
			}
			break;
 
			case SDL_QUIT:
			{
				m_bQuit = true;
			}
			break;

			case SDL_VIDEORESIZE: // It's a RESIZE Event?
			{

			}
			break;

 
			case SDL_ACTIVEEVENT:
			{
				if ( event.active.state & SDL_APPACTIVE ) 
				{
					if(event.active.gain ) 
					{
						if(m_bMinimized)
						{
							m_bMinimized = false;
							NotifyWindowActive();
						}
					}
					else
					{
						if(m_bMinimized == false)
						{
							m_bMinimized = true;
							NotifyWindowInactive();
						}
					}

					if(event.active.state & SDL_APPMOUSEFOCUS)					// The Mouse Cursor Has Left/Entered The Window Space?
					{
						if(event.active.gain)									// Entered?
							m_bMouseFocus = true;								// Report It Setting AppStatus.MouseFocus
						else													// Otherwhise
							m_bMouseFocus = false;								// The Cursor Has Left, Reset AppStatus.MouseFocus
					}

					if(event.active.state & SDL_APPINPUTFOCUS)					// The Window Has Gained/Lost Input Focus?
					{
						if(event.active.gain)									// Gained?
							m_bKeyboardFocus = true;							// Report It Where You Know (You Always Report, You're A Spy, Aren't You?!)
						else													// Otherwhise
							m_bKeyboardFocus = false;							// Reset AppStatus.KeyboardFocus
					}
				}
			}
			break;

		} // switch
	} // while (handling input)
}
 
/** Handles the updating routine. **/
bool Engine::DoUpdate() 
{
	//TODO: Create a framerate cap
	
	if( IsFPSCapEnabled() )
	{
		if( (SDL_GetTicks() - m_lLastTick) < (1000 / GetFPSCap() ) ) 
			SDL_Delay( (1000 / GetFPSCap()) - (SDL_GetTicks() - m_lLastTick) ); 
	}

	long iElapsedTicks = SDL_GetTicks() - m_lLastTick;
	m_lLastTick = SDL_GetTicks();

	m_fTimeDelta = (float)(iElapsedTicks);
	if(Update(m_fTimeDelta) == false)
		return false;
 
	m_nFPSTickCounter += iElapsedTicks;

	return true;
}
 
/** Handles the rendering and FPS calculations. **/
void Engine::DoRender()
{
	++m_nFPSCounter;
	if(m_nFPSTickCounter >= 1000 )
	{
		m_nCurrentFPS = m_nFPSCounter;
		m_nFPSCounter = 0;
		m_nFPSTickCounter = 0;
	}
	
	if(m_pScreen && m_pScreen->GetSDLSurface()) 
		Render( m_pScreen);	
}
 
bool Engine::SetVideoMode(int nWidth, int nHeight, bool bFullScreen, int nBpp, Uint32 nFlags)
{
	m_nWidth  = nWidth;
	m_nHeight = nHeight;
	m_bFullScreen = bFullScreen;
	m_nBpp = nBpp;

	if(m_bFullScreen)
		nFlags |= SDL_FULLSCREEN;

	printf("Engine: Checking video mode %ix%i@%dbpp.\n", m_nWidth, m_nHeight, m_nBpp);
	unsigned int bpp = SDL_VideoModeOK(m_nWidth, m_nHeight, m_nBpp, nFlags);
	if(!bpp)
	{
		fprintf( stderr, "Engine: Error, video mode not available (%s)\n", SDL_GetError() );
		return false;//exit(-1);
	}
	else
		printf("Engine: Video mode %ix%i@%dbpp supported\n", m_nWidth, m_nHeight, m_nBpp);

	//printf("SDL recomends mode %ix%i@%dbpp.\n", m_nWidth, m_nHeight, nBpp);
	//screen=SDL_SetVideoMode(640, 480, bpp, SDL_HWSURFACE);
			
	SDL_Surface *pScreen = SDL_SetVideoMode( m_nWidth, m_nHeight, m_nBpp, nFlags);
	if(!pScreen)
	{
		fprintf( stderr, "Engine: Error, unable to set SDL video mode (%s)\n", SDL_GetError() );
		return false;
	}
	else
	{	
		SDL_FillRect(pScreen, NULL, SDL_MapRGBA(pScreen->format,0,0,0,0));
		m_pScreen = new TextureSDL();
		if(m_pScreen->Initialize("present","", pScreen, true) == false)
		{
			m_pScreen->Release();
			m_pScreen = NULL;

			return false;
		}
	}

	return true;
}

void Engine::SetTitle(string sTitle)
{
	m_sTitle = sTitle;
	SDL_WM_SetCaption(m_sTitle.c_str(), 0 );
}